@@ -3,29 +3,29 @@
  * Author: Brent Hendricks
  * Project: GGZ Core Client Lib
  * Date: 6/5/00
- * $Id: room.c,v 1.68 2003/02/22 06:46:20 jdorje Exp $
+ * $Id: room.c 7889 2006-03-07 09:57:32Z josef $
  *
  * This fils contains functions for handling rooms
  *
  * Copyright (C) 1998 Brent Hendricks.
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ * 
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ * 
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #ifdef HAVE_CONFIG_H
-#  include <config.h>		/* Site-specific config */
+#  include <config.h>	/* Site-specific config */
 #endif
 
 #include <stdio.h>
@@ -44,13 +44,8 @@
 #include "server.h"
 #include "table.h"
 
-/* Local functions */
-static int _ggzcore_room_event_is_valid(GGZRoomEvent event);
-static GGZHookReturn _ggzcore_room_event(struct _GGZRoom *room, 
-					 GGZRoomEvent id, void *data);
-
 /* Array of GGZRoom messages */
-static char* _ggzcore_room_events[] = {
+static char *_ggzcore_room_events[] = {
 	"GGZ_PLAYER_LIST",
 	"GGZ_TABLE_LIST",
 	"GGZ_CHAT_EVENT",
@@ -64,20 +59,17 @@ static char* _ggzcore_room_events[] = {
 	"GGZ_TABLE_LEFT",
 	"GGZ_TABLE_LEAVE_FAIL",
 	"GGZ_PLAYER_LAG",
-	"GGZ_PLAYER_STATS"
+	"GGZ_PLAYER_STATS",
+	"GGZ_PLAYER_COUNT"
 };
 
-/* Total number of server events messages */
-static unsigned int _ggzcore_num_events = sizeof(_ggzcore_room_events)/sizeof(_ggzcore_room_events[0]);
-
-
 /*
- * The GGZRoom struct manages information about a particular 
+ * The GGZRoom struct manages information about a particular room
  */
 struct _GGZRoom {
 
 	/* Server which this room is on */
-	struct _GGZServer *server;
+	GGZServer *server;
 
 	/* Monitoring flag */
 	char monitor;
@@ -86,60 +78,217 @@ struct _GGZRoom {
 	unsigned int id;
 
 	/* Name of room */
-	char *name;	
-	
+	char *name;
+
 	/* Supported game type (ID on server) */
 	unsigned int game;
 
 	/* Room description */
 	char *desc;
 
-	/* Number of player */
+	/* Number of players in list room (current room only) */
 	unsigned int num_players;
 
-	/* List of players in the room */
+	/* List of players in the room (current room only) */
 	GGZList *players;
 
-	/* Number of tables */
+	/* Number of players we suspect are in the room */
+	int player_count;
+
+	/* Number of tables (current room only) */
 	unsigned int num_tables;
 
-	/* List of tables in the room */
+	/* List of tables in the room (current room only) */
 	GGZList *tables;
 
 	/* Room events */
-	GGZHookList *event_hooks[sizeof(_ggzcore_room_events)/sizeof(_ggzcore_room_events[0])];
+	GGZHookList *event_hooks[sizeof(_ggzcore_room_events) /
+				 sizeof(_ggzcore_room_events[0])];
 
 };
 
 
-/* Publicly exported functions */
-GGZRoom* ggzcore_room_new(void)
+/* Local functions */
+static int _ggzcore_room_event_is_valid(GGZRoomEvent event);
+
+/* Total number of server events messages */
+static unsigned int _ggzcore_num_events =
+    sizeof(_ggzcore_room_events) / sizeof(_ggzcore_room_events[0]);
+
+
+/*
+ * Private functions
+ */
+
+
+static GGZServer *_ggzcore_room_get_server(GGZRoom * room)
+{
+	return room->server;
+}
+
+
+static unsigned int _ggzcore_room_get_id(const GGZRoom * room)
+{
+	return room->id;
+}
+
+static const char *_ggzcore_room_get_name(GGZRoom * room)
+{
+	return room->name;
+}
+
+
+static struct _GGZGameType *_ggzcore_room_get_game(GGZRoom * room)
+{
+	return _ggzcore_server_get_type_by_id(room->server, room->game);
+}
+
+
+static const char *_ggzcore_room_get_desc(GGZRoom * room)
+{
+	return room->desc;
+}
+
+
+static unsigned int _ggzcore_room_get_num_players(GGZRoom * room)
+{
+	if (ggzcore_server_get_cur_room(room->server) == room) {
+		return room->num_players;
+	} else {
+		return room->player_count;
+	}
+}
+
+
+static GGZPlayer *_ggzcore_room_get_nth_player(GGZRoom * room,
+					       const unsigned int num)
+{
+	int i;
+	GGZListEntry *cur;
+
+	cur = ggz_list_head(room->players);
+	for (i = 0; i < num; i++)
+		cur = ggz_list_next(cur);
+
+	return ggz_list_get_data(cur);
+}
+
+
+GGZPlayer *_ggzcore_room_get_player_by_name(GGZRoom * room,
+					    const char *name)
+{
+	GGZListEntry *entry;
+	GGZPlayer *player, *found = NULL;
+
+	if (room->players) {
+		player = _ggzcore_player_new();
+		_ggzcore_player_init(player, name, NULL, -1, 0, 0);
+		entry = ggz_list_search(room->players, player);
+
+		if (entry)
+			found = ggz_list_get_data(entry);
+		_ggzcore_player_free(player);
+	}
+
+	return found;
+}
+
+
+static unsigned int _ggzcore_room_get_num_tables(GGZRoom * room)
+{
+	/* FIXME: we should let the list track this instead of doing it ourselves */
+	return room->num_tables;
+}
+
+
+static struct _GGZTable *_ggzcore_room_get_nth_table(GGZRoom * room,
+						     const unsigned int
+						     num)
+{
+	int i;
+	GGZListEntry *cur;
+
+	cur = ggz_list_head(room->tables);
+	for (i = 0; i < num; i++)
+		cur = ggz_list_next(cur);
+
+	return ggz_list_get_data(cur);
+}
+
+
+static struct _GGZTable *_ggzcore_room_get_table_by_id(GGZRoom * room,
+						       const unsigned int
+						       id)
+{
+	GGZTable *found = NULL;
+
+	if (room->tables) {
+		GGZListEntry *entry;
+		GGZTable *search_table = ggzcore_table_new();
+
+		_ggzcore_table_set_id(search_table, id);
+		entry = ggz_list_search(room->tables, search_table);
+
+		if (entry)
+			found = ggz_list_get_data(entry);
+
+		ggzcore_table_free(search_table);
+	}
+
+	return found;
+}
+
+
+/*
+ * Publicly exported functions
+ */
+
+
+GGZRoom *ggzcore_room_new(void)
 {
 	return _ggzcore_room_new();
 }
 
 
 /* Initialize room object */
-int ggzcore_room_init(GGZRoom *room, const GGZServer *server, const unsigned int id, const char *name, const unsigned int game, const char *desc)
+int ggzcore_room_init(GGZRoom * room, GGZServer * server,
+		      const unsigned int id, const char *name,
+		      const unsigned int game, const char *desc)
 {
 	if (room) {
-		_ggzcore_room_init(room, server, id, name, game, desc);
+		_ggzcore_room_init(room, server, id, name, game, desc, 0);
 		return 0;
-	}
-	else
+	} else
 		return -1;
 }
-		      
+
 
 /* De-allocate room object and its children */
-void ggzcore_room_free(GGZRoom *room)
+void ggzcore_room_free(GGZRoom * room)
 {
 	if (room)
 		_ggzcore_room_free(room);
 }
 
 
-char* ggzcore_room_get_name(GGZRoom *room)
+GGZServer *ggzcore_room_get_server(GGZRoom * room)
+{
+	if (room)
+		return _ggzcore_room_get_server(room);
+	else
+		return NULL;
+}
+
+
+int ggzcore_room_get_id(const GGZRoom * room)
+{
+	if (room)
+		return _ggzcore_room_get_id(room);
+	else
+		return -1;
+}
+
+const char *ggzcore_room_get_name(GGZRoom * room)
 {
 	if (room)
 		return _ggzcore_room_get_name(room);
@@ -148,7 +297,7 @@ char* ggzcore_room_get_name(GGZRoom *room)
 }
 
 
-char* ggzcore_room_get_desc(GGZRoom *room)
+const char *ggzcore_room_get_desc(GGZRoom * room)
 {
 	if (room)
 		return _ggzcore_room_get_desc(room);
@@ -157,7 +306,7 @@ char* ggzcore_room_get_desc(GGZRoom *room)
 }
 
 
-GGZGameType* ggzcore_room_get_gametype(GGZRoom *room)
+GGZGameType *ggzcore_room_get_gametype(GGZRoom * room)
 {
 	if (room)
 		return _ggzcore_room_get_game(room);
@@ -166,7 +315,7 @@ GGZGameType* ggzcore_room_get_gametype(GGZRoom *room)
 }
 
 
-int ggzcore_room_get_num_players(GGZRoom *room)
+int ggzcore_room_get_num_players(GGZRoom * room)
 {
 	if (room)
 		return _ggzcore_room_get_num_players(room);
@@ -175,7 +324,7 @@ int ggzcore_room_get_num_players(GGZRoom *room)
 }
 
 
-int ggzcore_room_get_num_tables(GGZRoom *room)
+int ggzcore_room_get_num_tables(GGZRoom * room)
 {
 	if (room)
 		return _ggzcore_room_get_num_tables(room);
@@ -184,7 +333,8 @@ int ggzcore_room_get_num_tables(GGZRoom *room)
 }
 
 
-GGZPlayer* ggzcore_room_get_nth_player(GGZRoom *room, const unsigned int num)
+GGZPlayer *ggzcore_room_get_nth_player(GGZRoom * room,
+				       const unsigned int num)
 {
 	if (room && num < room->num_players)
 		return _ggzcore_room_get_nth_player(room, num);
@@ -193,7 +343,8 @@ GGZPlayer* ggzcore_room_get_nth_player(GGZRoom *room, const unsigned int num)
 }
 
 
-GGZTable* ggzcore_room_get_nth_table(GGZRoom *room, const unsigned int num)
+GGZTable *ggzcore_room_get_nth_table(GGZRoom * room,
+				     const unsigned int num)
 {
 	if (room && num < room->num_tables)
 		return _ggzcore_room_get_nth_table(room, num);
@@ -203,7 +354,8 @@ GGZTable* ggzcore_room_get_nth_table(GGZRoom *room, const unsigned int num)
 
 
 
-GGZTable* ggzcore_room_get_table_by_id(GGZRoom *room, const unsigned int id)
+GGZTable *ggzcore_room_get_table_by_id(GGZRoom * room,
+				       const unsigned int id)
 {
 	if (!room)
 		return NULL;
@@ -213,25 +365,33 @@ GGZTable* ggzcore_room_get_table_by_id(GGZRoom *room, const unsigned int id)
 
 
 /* Functions for manipulating hooks to GGZRoom events */
-int ggzcore_room_add_event_hook(GGZRoom *room, const GGZRoomEvent event, const GGZHookFunc func)
+int ggzcore_room_add_event_hook(GGZRoom * room, const GGZRoomEvent event,
+				const GGZHookFunc func)
 {
 	if (room && func && _ggzcore_room_event_is_valid(event))
-		return _ggzcore_room_add_event_hook_full(room, event, func, NULL);
+		return _ggzcore_room_add_event_hook_full(room, event, func,
+							 NULL);
 	else
 		return -1;
 }
 
 
-int ggzcore_room_add_event_hook_full(GGZRoom *room, const GGZRoomEvent event, const GGZHookFunc func, void *data)
+int ggzcore_room_add_event_hook_full(GGZRoom * room,
+				     const GGZRoomEvent event,
+				     const GGZHookFunc func,
+				     const void *data)
 {
-	if (room && func  && _ggzcore_room_event_is_valid(event))
-		return _ggzcore_room_add_event_hook_full(room, event, func, data);
+	if (room && func && _ggzcore_room_event_is_valid(event))
+		return _ggzcore_room_add_event_hook_full(room, event, func,
+							 data);
 	else
 		return -1;
 }
 
 
-int ggzcore_room_remove_event_hook(GGZRoom *room, const GGZRoomEvent event, const GGZHookFunc func)
+int ggzcore_room_remove_event_hook(GGZRoom * room,
+				   const GGZRoomEvent event,
+				   const GGZHookFunc func)
 {
 	if (room && func && _ggzcore_room_event_is_valid(event))
 		return _ggzcore_room_remove_event_hook(room, event, func);
@@ -240,16 +400,19 @@ int ggzcore_room_remove_event_hook(GGZRoom *room, const GGZRoomEvent event, cons
 }
 
 
-int ggzcore_room_remove_event_hook_id(GGZRoom *room, const GGZRoomEvent event, const unsigned int hook_id)
+int ggzcore_room_remove_event_hook_id(GGZRoom * room,
+				      const GGZRoomEvent event,
+				      const unsigned int hook_id)
 {
 	if (room && _ggzcore_room_event_is_valid(event))
-		return _ggzcore_room_remove_event_hook_id(room, event, hook_id);
+		return _ggzcore_room_remove_event_hook_id(room, event,
+							  hook_id);
 	else
 		return -1;
 }
 
 
-int ggzcore_room_list_players(GGZRoom *room)
+int ggzcore_room_list_players(GGZRoom * room)
 {
 	if (room && room->server)
 		return _ggzcore_room_load_playerlist(room);
@@ -258,7 +421,8 @@ int ggzcore_room_list_players(GGZRoom *room)
 }
 
 
-int ggzcore_room_list_tables(GGZRoom *room, const int type, const char global)
+int ggzcore_room_list_tables(GGZRoom * room, const int type,
+			     const char global)
 {
 	if (room && room->server)
 		return _ggzcore_room_load_tablelist(room, type, global);
@@ -267,29 +431,28 @@ int ggzcore_room_list_tables(GGZRoom *room, const int type, const char global)
 }
 
 
-int ggzcore_room_chat(struct _GGZRoom *room, const GGZChatType type,
+int ggzcore_room_chat(GGZRoom * room, const GGZChatType type,
 		      const char *player, const char *msg)
 {
 	if (room && room->server) {
 		/* FIXME: check validty of args */
 		return _ggzcore_room_chat(room, type, player, msg);
-	}
-	else
+	} else
 		return -1;
 }
 
 
-int ggzcore_room_launch_table(GGZRoom *room, GGZTable *table)
+int ggzcore_room_launch_table(GGZRoom * room, GGZTable * table)
 {
 	if (room && room->server && table)
 		return _ggzcore_room_launch_table(room, table);
-	
+
 	else
 		return -1;
 }
 
 
-int ggzcore_room_join_table(GGZRoom *room, const unsigned int num,
+int ggzcore_room_join_table(GGZRoom * room, const unsigned int num,
 			    int spectator)
 {
 	if (!room
@@ -301,7 +464,7 @@ int ggzcore_room_join_table(GGZRoom *room, const unsigned int num,
 }
 
 
-int ggzcore_room_leave_table(GGZRoom *room, int force)
+int ggzcore_room_leave_table(GGZRoom * room, int force)
 {
 	if (room && room->server)
 		return _ggzcore_room_leave_table(room, force);
@@ -317,31 +480,32 @@ int ggzcore_room_leave_table(GGZRoom *room, int force)
 
 
 /* Create a new room object for a given server*/
-struct _GGZRoom* _ggzcore_room_new(void)
+GGZRoom *_ggzcore_room_new(void)
 {
-	struct _GGZRoom *room;
+	GGZRoom *room;
 
-	room = ggz_malloc(sizeof(struct _GGZRoom));
+	room = ggz_malloc(sizeof(GGZRoom));
 
 	return room;
 }
 
 
-void _ggzcore_room_init(struct _GGZRoom *room,
-			const struct _GGZServer *server,
-			const unsigned int id, 
-			const char* name, 
-			const unsigned int game, 
-			const char* desc)
+void _ggzcore_room_init(GGZRoom * room,
+			GGZServer * server,
+			const unsigned int id,
+			const char *name,
+			const unsigned int game,
+			const char *desc, const int player_count)
 {
 	int i;
-	
-	room->server = (struct _GGZServer*)server;
+
+	room->server = server;
 	room->id = id;
 	room->game = game;
 	room->name = ggz_strdup(name);
 	room->desc = ggz_strdup(desc);
-	
+	room->player_count = player_count;
+
 	/* FIXME: create player list? */
 	/* FIXME: create table list? */
 
@@ -351,10 +515,10 @@ void _ggzcore_room_init(struct _GGZRoom *room,
 }
 
 
-void _ggzcore_room_free(struct _GGZRoom *room)
+void _ggzcore_room_free(GGZRoom * room)
 {
 	int i;
-	
+
 	if (room->name)
 		ggz_free(room->name);
 
@@ -369,136 +533,51 @@ void _ggzcore_room_free(struct _GGZRoom *room)
 
 	for (i = 0; i < _ggzcore_num_events; i++)
 		_ggzcore_hook_list_destroy(room->event_hooks[i]);
-	
+
 	ggz_free(room);
 }
 
 
-/* Functions to retrieve room information */
-struct _GGZServer* _ggzcore_room_get_server(struct _GGZRoom *room)
+void _ggzcore_room_set_player_list(GGZRoom * room,
+				   unsigned int count, GGZList * list)
 {
-	return room->server;
-}
+	int count_changed = (count != room->num_players);
 
-
-unsigned int _ggzcore_room_get_id(struct _GGZRoom *room)
-{
-	return room->id;
-}
-
-
-char* _ggzcore_room_get_name(struct _GGZRoom *room)
-{
-	return room->name;
-}
-
-
-struct _GGZGameType* _ggzcore_room_get_game(struct _GGZRoom *room)
-{
-	return _ggzcore_server_get_type_by_id(room->server, room->game);
-}
-
-
-char* _ggzcore_room_get_desc(struct _GGZRoom *room)
-{
-	return room->desc;
-}
-
-
-unsigned int _ggzcore_room_get_num_players(struct _GGZRoom *room)
-{
-	return room->num_players;
-}
-
-
-struct _GGZPlayer* _ggzcore_room_get_nth_player(struct _GGZRoom *room, 
-						const unsigned int num)
-{
-	int i;
-	GGZListEntry *cur;
-	
-	cur = ggz_list_head(room->players);
-	for (i = 0; i < num; i++)
-		cur = ggz_list_next(cur);
-	
-	return ggz_list_get_data(cur);
-}
-
-
-struct _GGZPlayer* _ggzcore_room_get_player_by_name(struct _GGZRoom *room, 
-						    const char *name)
-{
-        GGZListEntry *entry;
-	struct _GGZPlayer player, *found = NULL;
-
-	if (room->players) {
-		player.name = (char*)name;
-		entry = ggz_list_search(room->players, &player);
-
-		if (entry)
-			found = ggz_list_get_data(entry);
-	}
-	
-	return found;
-}
-
-
-unsigned int _ggzcore_room_get_num_tables(struct _GGZRoom *room)
-{
-	/* FIXME: we should let the list track this instead of doing it ourselves */
-	return room->num_tables;
-}
-
-
-struct _GGZTable* _ggzcore_room_get_nth_table(struct _GGZRoom *room, 
-					       const unsigned int num)
-{
-	int i;
-	GGZListEntry *cur;
-	
-	cur = ggz_list_head(room->tables);
-	for (i = 0; i < num; i++)
-		cur = ggz_list_next(cur);
-	
-	return ggz_list_get_data(cur);
-}
-
-
-struct _GGZTable* _ggzcore_room_get_table_by_id(struct _GGZRoom *room, 
-						 const unsigned int id)
-{
-	GGZListEntry *entry;
-	struct _GGZTable table, *found = NULL;
-	
-	if (room->tables) {
-		_ggzcore_table_set_id(&table, id);
-		entry = ggz_list_search(room->tables, &table);
-		
-		if (entry)
-			found =  ggz_list_get_data(entry);
-	}
-
-	return found;
-}
-
-
-void _ggzcore_room_set_player_list(struct _GGZRoom *room,
-				   unsigned int count,
-				   GGZList *list)
-{
 	/* Get rid of old list */
 	ggz_list_free(room->players);
 
 	room->num_players = count;
+	room->player_count = count;
 	room->players = list;
 
-	_ggzcore_room_event(room, GGZ_PLAYER_LIST, NULL);
+	_ggzcore_room_event(room, GGZ_PLAYER_LIST, &room->id);
+	if (count_changed) {
+		GGZServer *server = _ggzcore_room_get_server(room);
+
+		_ggzcore_server_queue_players_changed(server);
+	}
 }
 
 
-void _ggzcore_room_set_table_list(struct _GGZRoom *room,
-				  unsigned int count,
-				  GGZList *list)
+void _ggzcore_room_set_players(GGZRoom * room, int players)
+{
+	GGZServer *server = _ggzcore_room_get_server(room);
+
+	if (room->player_count == players) {
+		return;
+	}
+	room->player_count = players;
+	if (room->player_count < 0) {
+		/* Sanity check. */
+		room->player_count = 0;
+	}
+	_ggzcore_room_event(room, GGZ_PLAYER_COUNT, &room->id);
+	_ggzcore_server_queue_players_changed(server);
+}
+
+
+void _ggzcore_room_set_table_list(GGZRoom * room,
+				  unsigned int count, GGZList * list)
 {
 	GGZListEntry *cur;
 	GGZTable *table;
@@ -514,12 +593,12 @@ void _ggzcore_room_set_table_list(struct _GGZRoom *room,
 		table = ggz_list_get_data(cur);
 		_ggzcore_table_set_room(table, room);
 	}
-	
+
 	_ggzcore_room_event(room, GGZ_TABLE_LIST, NULL);
 }
 
 
-void _ggzcore_room_set_monitor(struct _GGZRoom *room, char monitor)
+void _ggzcore_room_set_monitor(GGZRoom * room, char monitor)
 {
 	room->monitor = monitor;
 
@@ -536,184 +615,177 @@ void _ggzcore_room_set_monitor(struct _GGZRoom *room, char monitor)
 }
 
 
-void _ggzcore_room_add_player(struct _GGZRoom *room, GGZPlayer *pdata)
+void _ggzcore_room_add_player(GGZRoom * room, GGZPlayer * pdata,
+			      int from_room)
 {
-	struct _GGZPlayer *player;
+	GGZPlayer *player;
+	GGZRoomChangeEventData data;
+	GGZServer *server = _ggzcore_room_get_server(room);
+	int wins, losses, ties, forfeits, rating, ranking, highscore;
 
-	ggz_debug(GGZCORE_DBG_ROOM, "Adding player %s", pdata->name);
+	ggz_debug(GGZCORE_DBG_ROOM, "Adding player %s",
+		  ggzcore_player_get_name(pdata));
 
 	/* Create the list if it doesn't exist yet */
 	if (!room->players)
-		room->players = ggz_list_create(_ggzcore_player_compare, 
-						    NULL,
-						    _ggzcore_player_destroy, 0);
+		room->players = ggz_list_create(_ggzcore_player_compare,
+						NULL,
+						_ggzcore_player_destroy,
+						0);
+
+	_ggzcore_player_get_record(pdata, &wins, &losses, &ties,
+				   &forfeits);
+	_ggzcore_player_get_rating(pdata, &rating);
+	_ggzcore_player_get_ranking(pdata, &ranking);
+	_ggzcore_player_get_highscore(pdata, &highscore);
 
 	/* Default new people in room to no table (-1) */
 	player = _ggzcore_player_new();
-	_ggzcore_player_init(player, pdata->name, pdata->room,
-			     -1, pdata->type, pdata->lag);
-	_ggzcore_player_init_stats(player,
-				   pdata->wins,
-				   pdata->losses,
-				   pdata->ties,
-				   pdata->forfeits,
-				   pdata->rating,
-				   pdata->ranking,
-				   pdata->highscore);
+	_ggzcore_player_init(player, ggzcore_player_get_name(pdata),
+			     _ggzcore_player_get_room(pdata),
+			     -1, ggzcore_player_get_type(pdata),
+			     ggzcore_player_get_lag(pdata));
+	_ggzcore_player_init_stats(player, wins, losses, ties, forfeits,
+				   rating, ranking, highscore);
 
 	ggz_list_insert(room->players, player);
 	room->num_players++;
-	_ggzcore_room_event(room, GGZ_ROOM_ENTER, pdata->name);
-}
-			      
+	room->player_count = room->num_players;
 
-void _ggzcore_room_remove_player(struct _GGZRoom *room, char *name)
+	data.player_name = ggzcore_player_get_name(pdata);
+	data.from_room = from_room;
+	data.to_room = room->id;
+	_ggzcore_room_event(room, GGZ_ROOM_ENTER, &data);
+
+	if ((room =
+	     _ggzcore_server_get_room_by_id(room->server, from_room))) {
+		_ggzcore_room_set_players(room, room->player_count - 1);
+	}
+	_ggzcore_server_queue_players_changed(server);
+}
+
+
+void _ggzcore_room_remove_player(GGZRoom * room, const char *name,
+				 int to_room)
 {
-	struct _GGZPlayer player;
 	GGZListEntry *entry;
+	GGZRoomChangeEventData data;
 
 	ggz_debug(GGZCORE_DBG_ROOM, "Removing player %s", name);
 
 	/* Only try to delete if the list exists */
-	if (room->players) {	
-		player.name = name;
-		entry = ggz_list_search(room->players, &player);
+	if (room->players) {
+		GGZPlayer *player = _ggzcore_player_new();
+		_ggzcore_player_init(player, name, NULL, -1, 0, 0);
+		entry = ggz_list_search(room->players, player);
 		if (entry) {
+			GGZServer *server = _ggzcore_room_get_server(room);
+
 			ggz_list_delete_entry(room->players, entry);
 			room->num_players--;
-			_ggzcore_room_event(room, GGZ_ROOM_LEAVE, name);
+			room->player_count = room->num_players;
+
+			data.player_name = name;
+			data.from_room = room->id;
+			data.to_room = to_room;
+
+			_ggzcore_room_event(room, GGZ_ROOM_LEAVE, &data);
+			_ggzcore_server_queue_players_changed(server);
 		}
+		_ggzcore_player_free(player);
+	}
+
+	if ((room = _ggzcore_server_get_room_by_id(room->server, to_room))) {
+		_ggzcore_room_set_players(room, room->player_count + 1);
 	}
 }
 
 
-void _ggzcore_room_set_player_lag(struct _GGZRoom *room, char *name, int lag)
+void _ggzcore_room_add_table(GGZRoom * room, struct _GGZTable *table)
 {
-	/* FIXME: This should be sending a player "class-based" event */
-	struct _GGZPlayer *player;
+	ggz_debug(GGZCORE_DBG_ROOM, "Adding table %d",
+		  ggzcore_table_get_id(table));
 
-	ggz_debug(GGZCORE_DBG_ROOM, "Setting lag to %d for %s", lag, name);
-
-	if (room->players) {
-		player = _ggzcore_room_get_player_by_name(room, name);
-		if (player) /* make sure they're still in room */
-		{
-			_ggzcore_player_set_lag(player, lag);
-			_ggzcore_room_event(room, GGZ_PLAYER_LAG, name);
-		}
-	}
-}
-
-
-void _ggzcore_room_set_player_stats(GGZRoom *room, GGZPlayer *pdata)
-{
-	/* FIXME: This should be sending a player "class-based" event */
-	GGZPlayer *player;
-
-	ggz_debug(GGZCORE_DBG_ROOM, "Setting stats for %s: %d-%d-%d",
-		  pdata->name, pdata->wins, pdata->losses, pdata->ties);
-
-	if (!room->players)
-		return;
-
-	player = _ggzcore_room_get_player_by_name(room, pdata->name);
-
-	/* make sure they're still in room */
-	if (!player)
-		return;
-
-	_ggzcore_player_init_stats(player,
-				   pdata->wins,
-				   pdata->losses,
-				   pdata->ties,
-				   pdata->forfeits,
-				   pdata->rating,
-				   pdata->ranking,
-				   pdata->highscore);
-	_ggzcore_room_event(room, GGZ_PLAYER_STATS, player->name);
-}
-
-
-void _ggzcore_room_add_table(struct _GGZRoom *room, struct _GGZTable *table)
-{
-	ggz_debug(GGZCORE_DBG_ROOM, "Adding table %d", 
-		      _ggzcore_table_get_id(table));
-	
 	/* Set table to point to this room */
 	_ggzcore_table_set_room(table, room);
 
 	/* Create the list if it doesn't exist yet */
 	if (!room->tables)
-		room->tables = ggz_list_create(_ggzcore_table_compare, NULL,
-					       _ggzcore_table_destroy, 0);
-	
+		room->tables =
+		    ggz_list_create(_ggzcore_table_compare, NULL,
+				    _ggzcore_table_destroy, 0);
+
 	ggz_list_insert(room->tables, table);
 	room->num_tables++;
 	_ggzcore_room_event(room, GGZ_TABLE_UPDATE, NULL);
 }
 
 
-void _ggzcore_room_remove_table(struct _GGZRoom *room, const unsigned int id) 
+void _ggzcore_room_remove_table(GGZRoom * room, const unsigned int id)
 {
-	GGZListEntry *entry;
-	struct _GGZTable table;
-
 	ggz_debug(GGZCORE_DBG_ROOM, "Deleting table: %d", id);
 
 	/* Only try to delete if the list exists */
 	if (room->tables) {
-		_ggzcore_table_set_id(&table, id);
-		entry = ggz_list_search(room->tables, &table);
+		GGZListEntry *entry;
+		GGZTable *search_table = ggzcore_table_new();;
+
+		_ggzcore_table_set_id(search_table, id);
+
+		entry = ggz_list_search(room->tables, search_table);
 		if (entry) {
 			ggz_list_delete_entry(room->tables, entry);
 			room->num_tables--;
 			_ggzcore_room_event(room, GGZ_TABLE_UPDATE, NULL);
 		}
+
+		ggzcore_table_free(search_table);
 	}
 }
 
 
-void _ggzcore_room_player_set_table(struct _GGZRoom *room,
-				    const char * name, int table)
+void _ggzcore_room_player_set_table(GGZRoom * room,
+				    const char *name, int table)
 {
-	struct _GGZPlayer *player;
+	GGZPlayer *player;
 
 	ggz_debug(GGZCORE_DBG_ROOM, "%s table is now %d", name, table);
 
 	if (room->players) {
 		/* make sure they're still in room */
-		if ( (player = _ggzcore_room_get_player_by_name(room, name)))
+		if ((player =
+		     _ggzcore_room_get_player_by_name(room, name)))
 			_ggzcore_player_set_table(player, table);
 	}
-	
+
 	_ggzcore_room_event(room, GGZ_TABLE_UPDATE, NULL);
 }
 
 
-int _ggzcore_room_load_playerlist(struct _GGZRoom *room)
+int _ggzcore_room_load_playerlist(GGZRoom * room)
 {
-	struct _GGZNet *net;
+	GGZNet *net;
 
 	net = _ggzcore_server_get_net(room->server);
 	return _ggzcore_net_send_list_players(net);
 }
 
 
-int _ggzcore_room_load_tablelist(struct _GGZRoom *room, const int type, const char global)
+int _ggzcore_room_load_tablelist(GGZRoom * room, const int type,
+				 const char global)
 {
-	struct _GGZNet *net;
+	GGZNet *net;
 
 	net = _ggzcore_server_get_net(room->server);
 	return _ggzcore_net_send_list_tables(net, type, global);
 }
 
 
-int _ggzcore_room_chat(struct _GGZRoom *room,
+int _ggzcore_room_chat(GGZRoom * room,
 		       const GGZChatType type,
-		       const char *player,
-		       const char *msg)
+		       const char *player, const char *msg)
 {
-	struct _GGZNet *net;
+	GGZNet *net;
 
 	net = _ggzcore_server_get_net(room->server);
 
@@ -754,28 +826,34 @@ int _ggzcore_room_chat(struct _GGZRoom *room,
 }
 
 
-void _ggzcore_room_add_chat(struct _GGZRoom *room, GGZChatType type,
-			    char *name, char *msg)
+void _ggzcore_room_add_chat(GGZRoom * room, GGZChatType type,
+			    const char *name, const char *msg)
 {
-	GGZChatEventData data = { type : type,
-				  sender : name,
-				  message : msg };
+      GGZChatEventData data = { type:type,
+	      sender:name,
+	      message:msg
+	};
 
 	ggz_debug(GGZCORE_DBG_ROOM, "Chat (%s) from %s",
 		  ggz_chattype_to_string(type), name);
 
 	_ggzcore_room_event(room, GGZ_CHAT_EVENT, &data);
+	if (type == GGZ_CHAT_TABLE) {
+		GGZGame *game = _ggzcore_server_get_cur_game(room->server);
+
+		_ggzcore_game_inform_chat(game, name, msg);
+	}
 }
 
 
-void _ggzcore_room_set_table_launch_status(struct _GGZRoom *room, int status)
+void _ggzcore_room_set_table_launch_status(GGZRoom * room, int status)
 {
 	_ggzcore_server_set_table_launch_status(room->server, status);
 
 	if (status == E_OK) {
 		_ggzcore_room_event(room, GGZ_TABLE_LAUNCHED, NULL);
 	} else {
-		GGZErrorEventData error = {status: status};
+	      GGZErrorEventData error = { status:status };
 
 		switch (status) {
 		case E_BAD_OPTIONS:
@@ -816,22 +894,34 @@ void _ggzcore_room_set_table_launch_status(struct _GGZRoom *room, int status)
 }
 
 
-void _ggzcore_room_set_table_join(GGZRoom *room, int table_index)
+/* OK, we've joined a table.  Deal with it. */
+void _ggzcore_room_set_table_join(GGZRoom * room, int table_index)
 {
-	ggz_debug(GGZCORE_DBG_ROOM, "Player joined table %d.", table_index);
+	ggz_debug(GGZCORE_DBG_ROOM, "Player joined table %d.",
+		  table_index);
 	_ggzcore_server_set_table_join_status(room->server, E_OK);
 	_ggzcore_room_event(room, GGZ_TABLE_JOINED, &table_index);
-}
-					   
 
-void _ggzcore_room_set_table_join_status(struct _GGZRoom *room,
+	if (_ggzcore_server_get_cur_game(room->server) == NULL) {
+		/* GGZd thinks we're at a table but we know there's no game
+		 * client running.  Probably the game client exited during the
+		 * connection process.  So, tell ggzd that we've left. */
+		if (_ggzcore_room_leave_table(room, 1) < 0) {
+			/* Uh-oh; this shouldn't happen. */
+		}
+	}
+}
+
+
+void _ggzcore_room_set_table_join_status(GGZRoom * room,
 					 GGZClientReqError status)
 {
 	char buf[128];
 
 	if (status != E_OK)
-		_ggzcore_server_set_table_join_status(room->server, status);
-	
+		_ggzcore_server_set_table_join_status(room->server,
+						      status);
+
 	switch (status) {
 	case E_OK:
 		/* Do nothing if successful.  The join itself will
@@ -848,12 +938,12 @@ void _ggzcore_room_set_table_join_status(struct _GGZRoom *room,
 		_ggzcore_room_event(room, GGZ_TABLE_JOIN_FAIL,
 				    "Already at a table");
 		break;
-		
+
 	case E_IN_TRANSIT:
 		_ggzcore_room_event(room, GGZ_TABLE_JOIN_FAIL,
 				    "Already joining/leaving a table");
 		break;
-		
+
 	case E_BAD_OPTIONS:
 		_ggzcore_room_event(room, GGZ_TABLE_JOIN_FAIL,
 				    "Bad option");
@@ -863,7 +953,7 @@ void _ggzcore_room_set_table_join_status(struct _GGZRoom *room,
 		_ggzcore_room_event(room, GGZ_TABLE_JOIN_FAIL,
 				    "No such table");
 		break;
-		
+
 	case E_TABLE_FULL:
 		_ggzcore_room_event(room, GGZ_TABLE_JOIN_FAIL,
 				    "The table is full "
@@ -878,32 +968,33 @@ void _ggzcore_room_set_table_join_status(struct _GGZRoom *room,
 
 	default:
 		snprintf(buf, sizeof(buf),
-		         "Unknown join failure (status %d)",
-		         status);
+			 "Unknown join failure (status %d)", status);
 		_ggzcore_room_event(room, GGZ_TABLE_JOIN_FAIL, buf);
 		break;
 
 	}
 }
 
-void _ggzcore_room_set_table_leave(GGZRoom *room,
+void _ggzcore_room_set_table_leave(GGZRoom * room,
 				   GGZLeaveType reason, const char *player)
 {
-	GGZTableLeaveEventData event_data = {reason: reason,
-					     player: player};
+      GGZTableLeaveEventData event_data = { reason:reason,
+	      player:player
+	};
 	ggz_debug(GGZCORE_DBG_ROOM, "Player left table: %s (%s).",
 		  ggz_leavetype_to_string(reason), player);
 	_ggzcore_server_set_table_leave_status(room->server, E_OK);
 	_ggzcore_room_event(room, GGZ_TABLE_LEFT, &event_data);
 }
 
-void _ggzcore_room_set_table_leave_status(struct _GGZRoom *room,
+void _ggzcore_room_set_table_leave_status(GGZRoom * room,
 					  GGZClientReqError status)
 {
 	char buf[128];
 
 	if (status != E_OK)
-		_ggzcore_server_set_table_leave_status(room->server, status);
+		_ggzcore_server_set_table_leave_status(room->server,
+						       status);
 
 	switch (status) {
 	case E_OK:
@@ -921,7 +1012,7 @@ void _ggzcore_room_set_table_leave_status(struct _GGZRoom *room,
 		_ggzcore_room_event(room, GGZ_TABLE_LEAVE_FAIL,
 				    "No such table");
 		break;
-		
+
 	case E_LEAVE_FORBIDDEN:
 		_ggzcore_room_event(room, GGZ_TABLE_LEAVE_FAIL,
 				    "Cannot leave games of this type");
@@ -929,52 +1020,51 @@ void _ggzcore_room_set_table_leave_status(struct _GGZRoom *room,
 
 	default:
 		snprintf(buf, sizeof(buf),
-		         "Unknown leave failure (status %d)",
-		         status);
+			 "Unknown leave failure (status %d)", status);
 		_ggzcore_room_event(room, GGZ_TABLE_LEAVE_FAIL, buf);
 		break;
 	}
 }
 
 
-void _ggzcore_room_table_event(struct _GGZRoom *room, GGZRoomEvent event, void *data)
+void _ggzcore_room_table_event(GGZRoom * room, GGZRoomEvent event,
+			       void *data)
 {
 	_ggzcore_room_event(room, event, data);
 }
 
 
-int _ggzcore_room_launch_table(struct _GGZRoom *room, struct _GGZTable *table)
+int _ggzcore_room_launch_table(GGZRoom * room, struct _GGZTable *table)
 {
-	struct _GGZNet *net;
+	GGZNet *net;
 	int status;
 	GGZGame *game = ggzcore_server_get_cur_game(room->server);
 
 	/* Make sure we're actually in a room (FIXME: should probably
-           make sure we're in *this* room) and not already playing a
-           game */
+	   make sure we're in *this* room) and not already playing a
+	   game */
 	if (_ggzcore_server_get_state(room->server) != GGZ_STATE_IN_ROOM
 	    || !game)
 		return -1;
 
 	net = _ggzcore_server_get_net(room->server);
 	status = _ggzcore_net_send_table_launch(net, table);
-	
+
 	if (status == 0) {
 		_ggzcore_game_set_player(game, 0, -1);
 
 		_ggzcore_server_set_table_launching(room->server);
 	}
-	
+
 	return status;
 }
 
 
-int _ggzcore_room_join_table(struct _GGZRoom *room,
-			     const unsigned int num,
-			     int spectator)
+int _ggzcore_room_join_table(GGZRoom * room,
+			     const unsigned int num, int spectator)
 {
 	int status;
-	struct _GGZNet *net;
+	GGZNet *net;
 	GGZGame *game = ggzcore_server_get_cur_game(room->server);
 	GGZRoom *cur_room = _ggzcore_server_get_cur_room(room->server);
 	GGZTable *table;
@@ -1002,61 +1092,68 @@ int _ggzcore_room_join_table(struct _GGZRoom *room,
 }
 
 
-int _ggzcore_room_leave_table(struct _GGZRoom *room, int force)
+int _ggzcore_room_leave_table(GGZRoom * room, int force)
 {
 	int status;
-	struct _GGZNet *net;
+	GGZNet *net;
 	GGZGame *game = ggzcore_server_get_cur_game(room->server);
 	int spectating;
 
-	/* Make sure we're at a table (FIXME: should probably make
-           sure we're in *this* room) */
-	if (_ggzcore_server_get_state(room->server) != GGZ_STATE_AT_TABLE
-	    || !game)
+	/* Game may be NULL if the game client has already exited. */
+
+	/* Make sure we're at a table. (FIXME: should probably make
+	   sure we're in *this* room) */
+	if (_ggzcore_server_get_state(room->server) != GGZ_STATE_AT_TABLE) {
 		return -1;
+	}
 
 	net = _ggzcore_server_get_net(room->server);
-	spectating = _ggzcore_game_is_spectator(game);
+	if (game) {
+		spectating = _ggzcore_game_is_spectator(game);
+	} else {
+		spectating = 0;
+	}
 	status = _ggzcore_net_send_table_leave(net, force, spectating);
 
 	if (status == 0)
 		_ggzcore_server_set_table_leaving(room->server);
-	
+
 	return status;
 }
 
 
 /* Functions for attaching hooks to GGZRoom events */
-int _ggzcore_room_add_event_hook_full(GGZRoom *room,
-				      const GGZRoomEvent event, 
+int _ggzcore_room_add_event_hook_full(GGZRoom * room,
+				      const GGZRoomEvent event,
 				      const GGZHookFunc func,
-				      void *data)
+				      const void *data)
 {
-	return _ggzcore_hook_add_full(room->event_hooks[event], func, data);
+	return _ggzcore_hook_add_full(room->event_hooks[event], func,
+				      data);
 }
 
 
 /* Functions for removing hooks from GGZRoom events */
-int _ggzcore_room_remove_event_hook(GGZRoom *room,
-				    const GGZRoomEvent event, 
-				    const GGZHookFunc func) 
+int _ggzcore_room_remove_event_hook(GGZRoom * room,
+				    const GGZRoomEvent event,
+				    const GGZHookFunc func)
 {
 	return _ggzcore_hook_remove(room->event_hooks[event], func);
 }
 
 
-int _ggzcore_room_remove_event_hook_id(GGZRoom *room,
-				       const GGZRoomEvent event, 
-				       const unsigned int hook_id) 
+int _ggzcore_room_remove_event_hook_id(GGZRoom * room,
+				       const GGZRoomEvent event,
+				       const unsigned int hook_id)
 {
 	return _ggzcore_hook_remove_id(room->event_hooks[event], hook_id);
 }
 
 
 /* Return 0 if equal and -1 otherwise */
-int _ggzcore_room_compare(void* p, void* q)
+int _ggzcore_room_compare(void *p, void *q)
 {
-	if (((struct _GGZRoom*)p)->id == ((struct _GGZRoom*)q)->id)
+	if (((GGZRoom *) p)->id == ((GGZRoom *) q)->id)
 		return 0;
 
 	return -1;
@@ -1064,19 +1161,19 @@ int _ggzcore_room_compare(void* p, void* q)
 
 
 /* Create a copy of a room object */
-void* _ggzcore_room_copy(void* p)
+void *_ggzcore_room_copy(void *p)
 {
-	struct _GGZRoom *new, *src = p;
+	GGZRoom *new, *src = p;
 
 	new = _ggzcore_room_new();
 	_ggzcore_room_init(new, src->server, src->id, src->name, src->game,
-			   src->desc);
-	
-	return (void*)new;
+			   src->desc, src->player_count);
+
+	return (void *)new;
 }
 
 
-void  _ggzcore_room_destroy(void* p)
+void _ggzcore_room_destroy(void *p)
 {
 	_ggzcore_room_free(p);
 }
@@ -1092,8 +1189,8 @@ static int _ggzcore_room_event_is_valid(GGZRoomEvent event)
 }
 
 
-static GGZHookReturn _ggzcore_room_event(GGZRoom *room, GGZRoomEvent id, 
-					 void *data)
+GGZHookReturn _ggzcore_room_event(GGZRoom * room, GGZRoomEvent id,
+				  const void *data)
 {
 	return _ggzcore_hook_list_invoke(room->event_hooks[id], data);
 }
