@@ -4,28 +4,28 @@
  * Project: ggzmod
  * Date: 10/14/01
  * Desc: GGZ game module functions
- * $Id: ggzmod.c,v 1.9 2002/10/25 08:11:17 jdorje Exp $
+ * $Id: ggzmod.c 7888 2006-03-07 09:57:17Z josef $
  *
  * This file contains the backend for the ggzmod library.  This
- * library facilitates the communication between the GGZ server (ggz)
- * and game servers.  This file provides backend code that can be
+ * library facilitates the communication between the GGZ core client (ggz)
+ * and game clients.  This file provides backend code that can be
  * used at both ends.
  *
  * Copyright (C) 2001-2002 GGZ Development Team.
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ * 
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ * 
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #ifdef HAVE_CONFIG_H
@@ -39,17 +39,28 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <sys/select.h>
-#include <sys/socket.h>
+#ifdef HAVE_SYS_SELECT_H
+# include <sys/select.h>
+#endif
+#ifdef HAVE_SYS_SOCKET_H
+# include <sys/socket.h>
+#endif
+#ifdef HAVE_WINSOCK2_H
+# include <winsock2.h>
+#endif
 #include <sys/time.h>
 #include <sys/types.h>
-#include <sys/wait.h>
+#ifdef HAVE_SYS_WAIT_H
+# include <sys/wait.h>
+#endif
 #include <unistd.h>
 
 #include <ggz.h>
 
+#include "ggzcore-ggz.h"
+
 #include "ggzmod.h"
-#include "ggzmod-ggz.h"
+/*#include "ggzmod-ggz.h"*/
 #include "mod.h"
 #include "io.h"
 #include "protocol.h"
@@ -60,11 +71,8 @@
  */
 
 static void call_handler(GGZMod *ggzmod, GGZModEvent event, void *data);
-static void call_transaction(GGZMod * ggzmod, GGZModTransaction t, void *data);
 static int _ggzmod_handle_event(GGZMod * ggzmod, fd_set read_fds);
 static void _ggzmod_set_state(GGZMod * ggzmod, GGZModState state);
-static int send_game_launch(GGZMod * ggzmod);
-static int game_fork(GGZMod * ggzmod);
 
 /* Functions for manipulating seats */
 static GGZSeat* seat_copy(GGZSeat *orig);
@@ -78,6 +86,97 @@ static void spectator_seat_free(GGZSpectatorSeat *seat);
  * Creating/destroying a ggzmod object
  */
 
+static int stats_compare(const void *p, const void *q)
+{
+	const GGZStat *s_p = p, *s_q = q;
+
+	return s_p->number - s_q->number;
+}
+
+static void *stats_copy(void *porig)
+{
+	GGZStat *stat, *orig = porig;
+
+	stat = ggz_malloc(sizeof(*stat));
+	*stat = *orig;
+
+	return stat;
+}
+
+static void stats_free(void *pstat)
+{
+	GGZStat *stat = pstat;
+
+	ggz_free(stat);
+}
+
+static int infos_compare(const void *p, const void *q)
+{
+	const GGZPlayerInfo *s_p = p, *s_q = q;
+
+	return s_p->num - s_q->num;
+}
+
+static void *infos_copy(void *porig)
+{
+	GGZPlayerInfo *info, *orig = porig;
+
+	info = ggz_malloc(sizeof(*info));
+	*info = *orig;
+	info->realname = ggz_strdup(info->realname);
+	info->photo = ggz_strdup(info->photo);
+	info->host = ggz_strdup(info->host);
+
+	return info;
+}
+
+static void infos_free(void *pstat)
+{
+	GGZPlayerInfo *info = pstat;
+
+	ggz_free(info->realname);
+	ggz_free(info->photo);
+	ggz_free(info->host);
+	ggz_free(info);
+}
+
+/*
+ * How a game is launched (incomplete):
+ *
+ * 1.  ggz client invokes ggzcore to execute the game.  The game executable is
+ *     started.  If socketpair() is available then it is used to create a
+ *     ggz<->game socket connection.
+ *
+ * 2.  A game launch packet is sent from the GGZ client to the game client.
+ *
+ * 3.  The game client sets its state to CONNECTED (from created)..
+ *
+ * 4.  GGZ-client is informed of the game state change.
+ *
+ * 5a. ggzmod_set_server informs the game client about where to connect to.
+ *     A "server" packet is sent GGZ -> game.
+ *
+ * 6ab.The GGZ core client (or the game client) creates a socket and connects
+ *     to GGZ...
+ *
+ * 7b. ...GGZ_GAME_NEGOTIATED ggzcore event is triggered...
+ *
+ * 8b. ggzcore_set_server_fd is called by the GGZ core client...
+ *     ...it calls ggzmod_set_server_fd
+ *
+ * 9.  the game client sets its state to WAITING
+ *
+ * 10.  GGZ-client is informed of the game state change
+ *
+ * 11.  GGZ_GAME_PLAYING ggzcore event is triggered
+ *
+ * 12.  table-join or table-launch packet is sent from ggz CLIENT->SERVER
+ *
+ * 13.  ggz SERVER handles join/launch packets, sends response
+ *
+ * 14.  ... and then what? ...
+ */
+
 /* Creates a new ggzmod object. */
 GGZMod *ggzmod_new(GGZModType type)
 {
@@ -96,6 +195,9 @@ GGZMod *ggzmod_new(GGZModType type)
 	ggzmod->state = GGZMOD_STATE_CREATED;
 	ggzmod->fd = -1;
 	ggzmod->server_fd = -1;
+	ggzmod->server_host = NULL;
+	ggzmod->server_port = 0;
+	ggzmod->server_handle = NULL;
 	for (i = 0; i < GGZMOD_NUM_HANDLERS; i++)
 		ggzmod->handlers[i] = NULL;
 	ggzmod->gamedata = NULL;
@@ -112,10 +214,17 @@ GGZMod *ggzmod_new(GGZModType type)
 				GGZ_LIST_REPLACE_DUPS);
 	ggzmod->num_seats = ggzmod->num_spectator_seats = 0;
 
-	ggzmod->pid = -1;
-	ggzmod->argv = NULL;
-	for (i = 0; i < GGZMOD_NUM_TRANSACTIONS; i++)
-		ggzmod->thandlers[i] = NULL;
+	ggzmod->stats = ggz_list_create(stats_compare, stats_copy, stats_free,
+					GGZ_LIST_REPLACE_DUPS);
+	ggzmod->spectator_stats = ggz_list_create(stats_compare,
+						  stats_copy,
+						  stats_free,
+						  GGZ_LIST_REPLACE_DUPS);
+
+	ggzmod->infos = ggz_list_create(infos_compare, infos_copy, infos_free,
+					GGZ_LIST_REPLACE_DUPS);
+
+	/* GGZ-side only initialization code was here*/
 
 	return ggzmod;
 }
@@ -124,8 +233,6 @@ GGZMod *ggzmod_new(GGZModType type)
 /* Frees (deletes) a ggzmod object */
 void ggzmod_free(GGZMod * ggzmod)
 {
-	int i;
-
 	if (!ggzmod) {
 		return;
 	}
@@ -133,20 +240,15 @@ void ggzmod_free(GGZMod * ggzmod)
 	if (ggzmod->fd != -1)
 		(void)ggzmod_disconnect(ggzmod);
 	
+	if (ggzmod->server_host) ggz_free(ggzmod->server_host);
+	if (ggzmod->server_handle) ggz_free(ggzmod->server_handle);
+
 	ggzmod->type = -1;
 
 	if (ggzmod->my_name)
 		ggz_free(ggzmod->my_name);
 
-	if (ggzmod->pwd)
-		ggz_free(ggzmod->pwd);
-	
-	if (ggzmod->argv) {
-		for (i = 0; ggzmod->argv[i]; i++)
-			if (ggzmod->argv[i])
-				ggz_free(ggzmod->argv[i]);
-		ggz_free(ggzmod->argv);
-	}
+	/* GGZ-side only deinitialization code was here*/
 
 	/* Free the object */
 	ggz_free(ggzmod);
@@ -185,6 +287,15 @@ GGZModState ggzmod_get_state(GGZMod * ggzmod)
 }
 
 
+int ggzmod_get_server_fd(GGZMod *ggzmod)
+{
+	if (!ggzmod) {
+		return -1;
+	}
+	return ggzmod->server_fd;
+}
+
+
 void* ggzmod_get_gamedata(GGZMod * ggzmod)
 {
 	if (!ggzmod) {
@@ -194,39 +305,6 @@ void* ggzmod_get_gamedata(GGZMod * ggzmod)
 }
 
 
-void ggzmod_set_module(GGZMod * ggzmod, const char *pwd, char **argv)
-{
-	int i;
-
-	ggz_debug("GGZMOD", "Setting arguments");
-	
-	if (!ggzmod)
-		return;
-
-	if (ggzmod->type != GGZMOD_GGZ) {
-		_ggzmod_error(ggzmod, "Cannot set module args from module");
-		return;
-	}
-		
-	/* Check parameters */
-	if (!argv || !argv[0]) {
-		_ggzmod_error(ggzmod, "Bad module arguments");
-		return;
-	}
-
-	/* Count the number of args so we know how much to allocate */
-	for (i = 0; argv[i]; i++) {}
-
-	ggz_debug("GGZMOD", "Set %d arguments", i);
-	
-	ggzmod->argv = ggz_malloc(sizeof(char*)*(i+1));
-	ggzmod->pwd = ggz_strdup(pwd);
-	
-	for (i = 0; argv[i]; i++) 
-		ggzmod->argv[i] = ggz_strdup(argv[i]);
-}
-
-
 void ggzmod_set_gamedata(GGZMod * ggzmod, void * data)
 {
 	if (ggzmod)
@@ -234,18 +312,6 @@ void ggzmod_set_gamedata(GGZMod * ggzmod, void * data)
 }
 
 
-void ggzmod_set_server_fd(GGZMod * ggzmod, int fd)
-{
-	if (ggzmod && ggzmod->type == GGZMOD_GGZ) {
-		ggzmod->server_fd = fd;
-
-		/* If we're already connected, send the fd */
-		if (ggzmod->state == GGZMOD_STATE_WAITING)
-			_io_send_server(ggzmod->fd, fd);
-	}
-}
-
-
 void ggzmod_set_handler(GGZMod * ggzmod, GGZModEvent e,
 			 GGZModHandler func)
 {
@@ -259,27 +325,11 @@ void ggzmod_set_handler(GGZMod * ggzmod, GGZModEvent e,
 }
 
 
-void ggzmod_set_transaction_handler(GGZMod * ggzmod,
-				    GGZModTransaction t,
-				    GGZModTransactionHandler func)
-{
-	if (!ggzmod
-	    || t < 0 || t >= GGZMOD_NUM_TRANSACTIONS
-	    || ggzmod->type != GGZMOD_GGZ) {
-		ggz_error_msg("ggzmod_set_transaction_handler: "
-			      "invalid params");
-		return;
-	}
-
-	ggzmod->thandlers[t] = func;
-}
-
-
 int ggzmod_set_state(GGZMod * ggzmod, GGZModState state)
 {
 	if (!ggzmod)
 		return -1;
-	
+
 	if (ggzmod->type == GGZMOD_GAME) {
 		/* The game may only change the state from one of
 		   these two states. */
@@ -333,30 +383,22 @@ void _ggzmod_handle_player(GGZMod *ggzmod,
 const char *ggzmod_get_player(GGZMod *ggzmod,
 			      int *is_spectator, int *seat_num)
 {
-	if (ggzmod->state == GGZMOD_STATE_CREATED)
+	if (ggzmod->state == GGZMOD_STATE_CREATED) {
 		ggz_error_msg("ggzmod_get_my_seat:"
 			      " can't call when state is CREATED.");
+		return NULL;
+	}
 
-	*is_spectator = ggzmod->i_am_spectator;
-	*seat_num = ggzmod->my_seat_num;
+	if (is_spectator) {
+		*is_spectator = ggzmod->i_am_spectator;
+	}
+	if (seat_num) {
+		*seat_num = ggzmod->my_seat_num;
+	}
 
 	return ggzmod->my_name;
 }
 
-int ggzmod_set_player(GGZMod *ggzmod, const char *name,
-		      int is_spectator, int seat_num)
-{
-	if (!ggzmod
-	    || ggzmod->type != GGZMOD_GGZ)
-		return -1;
-
-	_ggzmod_set_player(ggzmod, name, is_spectator, seat_num);
-
-	if (ggzmod->state != GGZMOD_STATE_CREATED)
-		_io_send_player(ggzmod->fd, name, is_spectator, seat_num);
-
-	return 0;
-}
 
 /*
  * Seats and spectator seats.
@@ -491,36 +533,6 @@ void _ggzmod_handle_seat(GGZMod * ggzmod, GGZSeat *seat)
 	seat_free(old_seat);
 }
 
-int ggzmod_set_seat(GGZMod * ggzmod, GGZSeat *seat)
-{
-	GGZSeat oldseat;
-
-	if (ggzmod->type == GGZMOD_GAME)
-		return -1;
-
-	if (!seat || seat->num < 0) {
-		return -2;
-	}
-
-	/* If there is no such seat, return error */
-	oldseat = ggzmod_get_seat(ggzmod, seat->num);
-
-	if (oldseat.type == seat->type
-	    && ggz_strcmp(oldseat.name, seat->name) == 0) {
-		/* No change. */
-		return 0;
-	}
-
-	if (ggzmod->state != GGZMOD_STATE_CREATED) {
-		if (_io_send_seat(ggzmod->fd, seat) < 0)
-			_ggzmod_error(ggzmod, "Error writing to game");
-	}
-
-	_ggzmod_set_seat(ggzmod, seat);
-
-	return 0;
-}
-
 static void _ggzmod_set_spectator_seat(GGZMod * ggzmod, GGZSpectatorSeat *seat)
 {
 	if (seat->name) {
@@ -565,51 +577,116 @@ void _ggzmod_handle_spectator_seat(GGZMod * ggzmod, GGZSpectatorSeat * seat)
 	spectator_seat_free(old_seat);
 }
 
-int ggzmod_set_spectator_seat(GGZMod * ggzmod, GGZSpectatorSeat *seat)
+void _ggzmod_handle_chat(GGZMod *ggzmod, char *player, char *chat_msg)
 {
-	if (!seat) return -1;
-	if (ggzmod->type == GGZMOD_GAME) return -2;
-	if (seat->num < 0) return -3;
+	GGZChat chat;
 
-	if (ggzmod->state != GGZMOD_STATE_CREATED) {
-		GGZSpectatorSeat old_seat;
-		old_seat = ggzmod_get_spectator_seat(ggzmod, seat->num);
-		if (ggz_strcmp(seat->name, old_seat.name)
-		    && _io_send_spectator_seat(ggzmod->fd, seat) < 0) {
-			_ggzmod_error(ggzmod, "Error writing to game");
-			return -4;
+	chat.player = player;
+	chat.message = chat_msg;
+
+	call_handler(ggzmod, GGZMOD_EVENT_CHAT, &chat);
+}
+
+void _ggzmod_handle_stats(GGZMod *ggzmod, GGZStat *player_stats,
+			  GGZStat *spectator_stats)
+{
+	GGZListEntry *entry;
+	GGZSeat *seat;
+	GGZSpectatorSeat *spectator;
+	GGZStat stat;
+	int i;
+
+	i = 0;
+	for(entry = ggz_list_head(ggzmod->seats); entry; entry = ggz_list_next(entry)) {
+		seat = ggz_list_get_data(entry);
+		stat = player_stats[i];
+		stat.number = seat->num;
+		ggz_list_insert(ggzmod->stats, &stat);
+		i++;
+	}
+
+	i = 0;
+	for(entry = ggz_list_head(ggzmod->spectator_seats); entry; entry = ggz_list_next(entry)) {
+		spectator = ggz_list_get_data(entry);
+		stat = spectator_stats[i];
+		stat.number = spectator->num;
+		ggz_list_insert(ggzmod->spectator_stats, &stat);
+		i++;
+	}
+
+	call_handler(ggzmod, GGZMOD_EVENT_STATS, NULL);
+}
+
+void _ggzmod_handle_info(GGZMod * ggzmod, int seat_num, const char *realname,
+			 const char *photo, const char *host, int finish)
+{
+	GGZPlayerInfo info = {
+		.num = seat_num,
+		.realname = realname,
+		.photo = photo,
+		.host = host
+	};
+
+	if(seat_num != -1) {
+		ggz_list_insert(ggzmod->infos, &info);
+	}
+
+	if(finish) {
+		if(seat_num == -1) {
+			call_handler(ggzmod, GGZMOD_EVENT_INFO, NULL);
+		} else {
+			call_handler(ggzmod, GGZMOD_EVENT_INFO, &info);
 		}
 	}
 
-	_ggzmod_set_spectator_seat(ggzmod, seat);
-
-	return 0;
 }
 
+
 /* 
  * GGZmod actions
  */
 
 int ggzmod_connect(GGZMod * ggzmod)
 {
+	char *ggzsocketstr;
+	int ggzsocket;
+	int items;
+
 	if (!ggzmod)
 		return -1;
 
-	if (ggzmod->type == GGZMOD_GGZ) {
-		/* For the ggz side, we fork the game and then send the launch message */
-		
-		if (game_fork(ggzmod) < 0) {
-			_ggzmod_error(ggzmod, "Error: table fork failed");
-			return -1;
-		}
-		
-		if (send_game_launch(ggzmod) < 0) {
-			_ggzmod_error(ggzmod, "Error sending launch to game");
-			return -1;
+	if (ggzmod->type == GGZMOD_GAME) {
+		ggzsocket = 0;
+#ifdef HAVE_GETENV
+		ggzsocketstr = getenv("GGZSOCKET");
+#else
+		GetEnvironmentVariable("GGZSOCKET", buf, sizeof(buf));
+#endif
+		if(ggzsocketstr) {
+			items = sscanf(ggzsocketstr, "%d", &ggzsocket);
+			if (items == 0) {
+				ggzsocket = 103;
+			}
+		} else {
+			ggzsocket = 103;
 		}
 
-	} else
-		ggzmod->fd = 3;
+#ifdef HAVE_SOCKETPAIR
+		ggzmod->fd = ggzsocket;
+#else /* Winsock implementation: see game_fork(). */
+		int sock;
+
+		if (ggzsocket == 0) {
+			ggz_error_msg_exit("Could not determine port.");
+		}
+
+		sock = ggz_make_socket(GGZ_SOCK_CLIENT, ggzsocket, "localhost");
+		if (sock < 0) {
+			ggz_error_msg_exit("Could not connect to port.");
+		}
+		ggzmod->fd = sock;
+#endif
+	}
 	
 	return 0;
 }
@@ -631,54 +708,52 @@ int ggzmod_dispatch(GGZMod * ggzmod)
 	FD_SET(ggzmod->fd, &read_fd_set);
 
 	timeout.tv_sec = timeout.tv_usec = 0;	/* is this really portable? */
-	
+
 	status = select(ggzmod->fd + 1, &read_fd_set, NULL, NULL, &timeout);
-	
-	if (status <= 0) {
+
+	if (status == 0) {
+		/* Nothing to read. */
+		return 0;
+	} else if (status < 0) {
 		if (errno == EINTR)
 			return 0;
 		return -1;
 	}
-	
+
 	return _ggzmod_handle_event(ggzmod, read_fd_set);
 }
 
-
 int ggzmod_disconnect(GGZMod * ggzmod)
 {
-	if (!ggzmod || ggzmod->fd < 0 ) {
+	if (!ggzmod) {
 		return -1;
 	}
+	if (ggzmod->fd < 0) {
+		/* This isn't an error; it usually means
+		   we already disconnected.  The invariant is that the
+		   process (ggzmod->pid) exists iff the socket (ggzmod->fd)
+		   exists. */
+		return 0;
+	}
 
-	if (ggzmod->type == GGZMOD_GGZ) {
-		/* For the ggz side, we kill the game server and close the socket */
-		
-		/* Make sure game server is dead */
-		if (ggzmod->pid > 0) {
-			kill(ggzmod->pid, SIGINT);
-			/* This will block waiting for the child to exit.
-			   This could be a problem if there is an error
-			   (or if the child refuses to exit...). */
-			(void) waitpid(ggzmod->pid, NULL, 0);
-		}
-		ggzmod->pid = -1;
-		
-		_ggzmod_set_state(ggzmod, GGZMOD_STATE_DONE);
-		/* FIXME: what other cleanups should we do? */
-	} else {
+	if (ggzmod->type == GGZMOD_GAME) {
 		/* For client the game side we send a game over message */
-		
+
 		/* First warn the server of halt (if we haven't already) */
 		_ggzmod_set_state(ggzmod, GGZMOD_STATE_DONE);
 		ggz_debug("GGZMOD", "Disconnected from GGZ server.");
 	}
-	
+
 	/* We no longer free the seat data here.  It will stick around until
 	   ggzmod_free is called or it is used again. */
 
 	/* Clean up the ggzmod object.  In theory it could now reconnect for
 	   a new game. */
+#ifdef HAVE_WINSOCK2_H
+	closesocket(ggzmod->fd);
+#else
 	close(ggzmod->fd);
+#endif
 	ggzmod->fd = -1;
 
 	return 0;
@@ -726,114 +801,6 @@ static void _ggzmod_set_state(GGZMod * ggzmod, GGZModState state)
 }
 
 
-
-
-/* 
- * ggz specific actions
- */
-
-/* Sends a game launch packet to ggzmod-game. A negative return value
-   indicates a serious (fatal) error. */
-static int send_game_launch(GGZMod * ggzmod)
-{
-	GGZListEntry *entry;
-
-	if (_io_send_player(ggzmod->fd,
-			    ggzmod->my_name,
-			    ggzmod->i_am_spectator,
-			    ggzmod->my_seat_num) < 0)
-		return -2;
-
-	for (entry = ggz_list_head(ggzmod->seats);
-	     entry;
-	     entry = ggz_list_next(entry)) {
-		GGZSeat *seat = ggz_list_get_data(entry);
-		if (_io_send_seat(ggzmod->fd, seat) < 0)
-			return -3;
-	}
-	for (entry = ggz_list_head(ggzmod->spectator_seats);
-	     entry;
-	     entry = ggz_list_next(entry)) {
-		GGZSpectatorSeat *seat = ggz_list_get_data(entry);
-		if (_io_send_spectator_seat(ggzmod->fd, seat) < 0)
-			return -4;
-	}
-
-	if (_io_send_launch(ggzmod->fd) < 0)
-		return -1;
-
-	/* If the server fd has already been set, send that too */
-	if (ggzmod->server_fd != -1)
-		if (_io_send_server(ggzmod->fd, ggzmod->server_fd) < 0)
-			return -5;
-
-	return 0;
-}
-
-
-/* Forks the game.  A negative return value indicates a serious error. */
-/* No locking should be necessary within this function. */
-static int game_fork(GGZMod * ggzmod)
-{
-	int pid;
-	int fd_pair[2];		/* socketpair */
-
-	/* If there are no args, we don't know what to run! */
-	if (ggzmod->argv == NULL || ggzmod->argv[0] == NULL) {
-		_ggzmod_error(ggzmod, "No arguments");
-		return -1;
-	}
-
-	if (socketpair(PF_LOCAL, SOCK_STREAM, 0, fd_pair) < 0)
-		ggz_error_sys_exit("socketpair failed");
-
-	if ( (pid = fork()) < 0)
-		ggz_error_sys_exit("fork failed");
-	else if (pid == 0) {
-		/* child */
-		close(fd_pair[0]);
-
-		/* debugging message??? */
-
-		/* Now we copy one end of the socketpair to fd 3 */
-		if (fd_pair[1] != 3) {
-			/* We'd like to send an error message if either of
-			   these fail, but we can't.  --JDS */
-			if (dup2(fd_pair[1], 3) != 3 || close(fd_pair[1]) < 0)
-				ggz_error_sys_exit("dup failed");
-		}
-
-		/* FIXME: Close all other fd's? */
-		/* FIXME: Not necessary to close other fd's if we use
-		   CLOSE_ON_EXEC */
-
-		/* Set working directory */
-		if (ggzmod->pwd
-		    && chdir(ggzmod->pwd) < 0) {
-			/* FIXME: what to do? */
-		}
-
-		/* FIXME: can we call ggzmod_log() from here? */
-		execv(ggzmod->argv[0], ggzmod->argv);	/* run game */
-
-		/* We should never get here.  If we do, it's an eror */
-		/* we still can't send error messages... */
-		ggz_error_sys_exit("exec of %s failed", ggzmod->argv[0]);
-	} else {
-		/* parent */
-		close(fd_pair[1]);
-
-		ggzmod->fd = fd_pair[0];
-		ggzmod->pid = pid;
-		
-		/* FIXME: should we delete the argv arguments? */
-		
-		/* That's all! */
-	}
-	return 0;
-}
-
-
 /**** Transaction requests  ****/
 
 void ggzmod_request_stand(GGZMod * ggzmod)
@@ -861,6 +828,11 @@ void ggzmod_request_open(GGZMod * ggzmod, int seat_num)
 	_io_send_request_open(ggzmod->fd, seat_num);
 }
 
+void ggzmod_request_chat(GGZMod *ggzmod, const char *chat_msg)
+{
+	_io_send_request_chat(ggzmod->fd, chat_msg);
+}
+
 /**** Internal library functions ****/
 
 /* Invokes handlers for the specified event */
@@ -871,22 +843,6 @@ static void call_handler(GGZMod *ggzmod, GGZModEvent event, void *data)
 }
 
 
-static void call_transaction(GGZMod * ggzmod, GGZModTransaction t, void *data)
-{
-	if (!ggzmod->thandlers[t]) {
-		ggz_error_msg("Unhandled transaction %d.", t);
-		return;
-	}
-
-	if (ggzmod->type != GGZMOD_GGZ) {
-		ggz_error_msg("The game can't handle transactions!");
-		return;
-	}
-
-	(*ggzmod->thandlers[t])(ggzmod, t, data);
-}
-
-
 void _ggzmod_error(GGZMod *ggzmod, char* error)
 {
 	call_handler(ggzmod, GGZMOD_EVENT_ERROR, error);
@@ -899,15 +855,18 @@ void _ggzmod_handle_state(GGZMod * ggzmod, GGZModState state)
 	   and they can only change it if the state is currently
 	   WAITING or PLAYING. */
 	switch (state) {
+	case GGZMOD_STATE_CREATED:
+	case GGZMOD_STATE_CONNECTED:
 	case GGZMOD_STATE_WAITING:
 	case GGZMOD_STATE_PLAYING:
 	case GGZMOD_STATE_DONE:
 		/* In contradiction to what I say above, the game
 		   actually _is_ allowed to change its state from
-		   CREATED to WAITING.  When ggzmod-ggz sends a
+		   CREATED to CONNECTED and subsequently to WAITING.
+		   or PLAYING.  When ggzmod-ggz sends a
 		   launch packet to ggzmod-game, ggzmod-game
 		   automatically changes the state from CREATED
-		   to WAITING.  When this happens, it tells
+		   to CONNECTED.  When this happens, it tells
 		   ggzmod-ggz of this change and we end up back
 		   here.  So, although it's a bit unsafe, we have
 		   to allow this for now.  The alternative would
@@ -915,51 +874,174 @@ void _ggzmod_handle_state(GGZMod * ggzmod, GGZModState state)
 		   separately change states when the launch packet
 		   is sent. */
 		_ggzmod_set_state(ggzmod, state);
-		break;
-	default:
-		_ggzmod_error(ggzmod,
-			       "Game requested incorrect state value");
+		return;
 	}
-	
-	/* Is this right? has the gameover happened yet? */   
-}
+	_ggzmod_error(ggzmod,
+		      "Game requested incorrect state value");
 
-void _ggzmod_handle_stand_request(GGZMod *ggzmod)
-{
-	call_transaction(ggzmod, GGZMOD_TRANSACTION_STAND, NULL);
-}
-
-void _ggzmod_handle_sit_request(GGZMod *ggzmod, int seat_num)
-{
-	call_transaction(ggzmod, GGZMOD_TRANSACTION_SIT, &seat_num);
-}
-
-void _ggzmod_handle_boot_request(GGZMod *ggzmod, char *name)
-{
-	call_transaction(ggzmod, GGZMOD_TRANSACTION_BOOT, name);
-}
-
-void _ggzmod_handle_bot_request(GGZMod *ggzmod, int seat_num)
-{
-	call_transaction(ggzmod, GGZMOD_TRANSACTION_BOT, &seat_num);
-}
-
-void _ggzmod_handle_open_request(GGZMod *ggzmod, int seat_num)
-{
-	call_transaction(ggzmod, GGZMOD_TRANSACTION_OPEN, &seat_num);
+	/* Is this right? has the gameover happened yet? */
 }
 
 
 void _ggzmod_handle_launch(GGZMod * ggzmod)
 {
 	/* Normally we let the game control its own state, but
-	   we control the transition from CREATED to WAITING. */
-	_ggzmod_set_state(ggzmod, GGZMOD_STATE_WAITING);
+	   we control the transition from CREATED to CONNECTED. */
+	_ggzmod_set_state(ggzmod, GGZMOD_STATE_CONNECTED);
 }
 
 
-void _ggzmod_handle_server(GGZMod * ggzmod, int fd)
+void _ggzmod_handle_server(GGZMod * ggzmod, const char *host,
+			   unsigned int port, const char *handle)
+{
+	ggzmod->server_host = ggz_strdup(host);
+	ggzmod->server_port = port;
+	ggzmod->server_handle = ggz_strdup(handle);
+
+	/* In the case of game clients connecting themselves, do it here. */
+	ggzmod->server_fd = ggzcore_channel_connect(host, port, handle);
+	if (ggzmod->server_fd < 0) {
+		_ggzmod_error(ggzmod, "Could not create channel.");
+		return;
+	}
+	_ggzmod_set_state(ggzmod, GGZMOD_STATE_WAITING);
+	call_handler(ggzmod, GGZMOD_EVENT_SERVER, &ggzmod->server_fd);
+}
+
+
+void _ggzmod_handle_server_fd(GGZMod * ggzmod, int fd)
 {
 	ggzmod->server_fd = fd;
-	call_handler(ggzmod, GGZMOD_EVENT_SERVER, &fd);
+
+	/* Everything's done now, so we move into the waiting state. */
+	_ggzmod_set_state(ggzmod, GGZMOD_STATE_WAITING);
+	call_handler(ggzmod, GGZMOD_EVENT_SERVER, &ggzmod->server_fd);
+}
+
+
+int ggzmod_player_get_record(GGZMod *ggzmod, GGZSeat *seat,
+			     int *wins, int *losses,
+			     int *ties, int *forfeits)
+{
+	GGZStat search_stat = {.number = seat->num};
+	GGZListEntry *entry = ggz_list_search(ggzmod->stats, &search_stat);
+	GGZStat *stat = ggz_list_get_data(entry);
+
+	if (!stat || !stat->have_record) return 0;
+
+	*wins = stat->wins;
+	*losses = stat->losses;
+	*ties = stat->ties;
+	*forfeits = stat->forfeits;
+	return 1;
+}
+
+int ggzmod_player_get_rating(GGZMod *ggzmod, GGZSeat *seat, int *rating)
+{
+	GGZStat search_stat = {.number = seat->num};
+	GGZListEntry *entry = ggz_list_search(ggzmod->stats, &search_stat);
+	GGZStat *stat = ggz_list_get_data(entry);
+
+	if (!stat || !stat->have_rating) return 0;
+
+	*rating = stat->rating;
+	return 1;
+}
+
+int ggzmod_player_get_ranking(GGZMod *ggzmod, GGZSeat *seat, int *ranking)
+{
+	GGZStat search_stat = {.number = seat->num};
+	GGZListEntry *entry = ggz_list_search(ggzmod->stats, &search_stat);
+	GGZStat *stat = ggz_list_get_data(entry);
+
+	if (!stat || !stat->have_ranking) return 0;
+
+	*ranking = stat->ranking;
+	return 1;
+}
+
+int ggzmod_player_get_highscore(GGZMod *ggzmod, GGZSeat *seat, int *highscore)
+{
+	GGZStat search_stat = {.number = seat->num};
+	GGZListEntry *entry = ggz_list_search(ggzmod->stats, &search_stat);
+	GGZStat *stat = ggz_list_get_data(entry);
+
+	if (!stat || !stat->have_highscore) return 0;
+
+	*highscore = stat->highscore;
+	return 1;
+}
+
+int ggzmod_spectator_get_record(GGZMod *ggzmod, GGZSpectatorSeat *seat,
+				int *wins, int *losses,
+				int *ties, int *forfeits)
+{
+	GGZStat search_stat = {.number = seat->num};
+	GGZListEntry *entry = ggz_list_search(ggzmod->spectator_stats,
+					      &search_stat);
+	GGZStat *stat = ggz_list_get_data(entry);
+
+	if(!stat) return 0;
+	*wins = stat->wins;
+	*losses = stat->losses;
+	*ties = stat->ties;
+	*forfeits = stat->forfeits;
+	return 1;
+}
+
+int ggzmod_spectator_get_rating(GGZMod *ggzmod, GGZSpectatorSeat *seat, int *rating)
+{
+	GGZStat search_stat = {.number = seat->num};
+	GGZListEntry *entry = ggz_list_search(ggzmod->spectator_stats,
+					      &search_stat);
+	GGZStat *stat = ggz_list_get_data(entry);
+
+	if(!stat) return 0;
+	*rating = stat->rating;
+	return 1;
+}
+
+int ggzmod_spectator_get_ranking(GGZMod *ggzmod, GGZSpectatorSeat *seat, int *ranking)
+{
+	GGZStat search_stat = {.number = seat->num};
+	GGZListEntry *entry = ggz_list_search(ggzmod->spectator_stats,
+					      &search_stat);
+	GGZStat *stat = ggz_list_get_data(entry);
+
+	if(!stat) return 0;
+	*ranking = stat->ranking;
+	return 1;
+}
+
+int ggzmod_spectator_get_highscore(GGZMod *ggzmod, GGZSpectatorSeat *seat, int *highscore)
+{
+	GGZStat search_stat = {.number = seat->num};
+	GGZListEntry *entry = ggz_list_search(ggzmod->spectator_stats,
+					      &search_stat);
+	GGZStat *stat = ggz_list_get_data(entry);
+
+	if(!stat) return 0;
+	*highscore = stat->highscore;
+	return 1;
+}
+
+int ggzmod_player_request_info(GGZMod *ggzmod, int seat_num)
+{
+	GGZSeat seat;
+	if(seat_num != -1) {
+		if(seat_num < -1) return 0;
+		if(seat_num >= ggzmod_get_num_seats(ggzmod)) return 0;
+		seat = ggzmod_get_seat(ggzmod, seat_num);
+		if(seat.type != GGZ_SEAT_PLAYER) return 0;
+	}
+	_io_send_req_info(ggzmod->fd, seat_num);
+	return 1;
+}
+
+GGZPlayerInfo* ggzmod_player_get_info(GGZMod *ggzmod, int seat)
+{
+	GGZPlayerInfo search_info = {.num = seat};
+	GGZListEntry *entry = ggz_list_search(ggzmod->infos, &search_info);
+
+	return entry ?  ggz_list_get_data(entry) : NULL;
 }
